# -*- coding: utf-8 -*-
"""
Created on Sun Jun 14 15:13:29 2020

@Author: Zhi-Jiang Yang, Dong-Sheng Cao
@Institution: CBDD Group, Xiangya School of Pharmaceutical Science, CSU, China
@Homepage: http://www.scbdd.com
@Mail: yzjkid9@gmail.com; oriental-cds@163.com
@Blog: https://blog.iamkotori.com

♥I love Princess Zelda forever♥
"""


from rdkit import Chem
from rdkit.Chem.rdmolops import RDKFingerprint
from rdkit.Chem.rdmolops import UnfoldedRDKFingerprintCountBased
from rdkit.Chem.Draw import rdMolDraw2D
from rdkit.Chem.Draw import _getRDKitEnv

__all__ = ['GetFoldedPathFragment',
           'GetUnfoldedPathFragment',
           'GetPathFragment',
           'DrawRDKitEnv']


def _DisposePathFragments(bitInfo, maxFragment=True):
    """Dispose the bitinfo retrived from GetFoldedPathFragment() or GetUnfoldedPathFragment()
    *internal only*
    """
    def func(frags):
        sets = [set(x) for e in frags.values() for x in e]
        us = []
        for e in sets:
            if any(e < s for s in sets):
                continue
            else:
                us.append(str(e))
        return us

    maxFragments = []
    allFragments = list(bitInfo.keys())
    if maxFragment:
        dic = {str(set(x)): k for k, v in bitInfo.items() for x in v}
        bitInfo = func(bitInfo)
        for k in bitInfo:
            idx = dic[k]
            if idx not in maxFragments:
                maxFragments.append(idx)
    else:
        pass

    return (allFragments, maxFragments)


def DrawRDKitEnv(mol, bondPath, molSize=(150, 150), baseRad=0.3, useSVG=True,
                 aromaticColor=(0.9, 0.9, 0.2), extraColor=(0.9, 0.9, 0.9), nonAromaticColor=None,
                 drawOptions=None, **kwargs):
    """Get SMARTS and SVG image from given bonds
    *internal only*
    """
    menv = _getRDKitEnv(mol, bondPath, baseRad, aromaticColor,
                        extraColor, nonAromaticColor, **kwargs)
    submol = menv.submol
    subMol = Chem.MolToSmiles(submol, isomericSmiles=False)

    # Drawing
    if useSVG:
        drawer = rdMolDraw2D.MolDraw2DSVG(molSize[0], molSize[1])
    else:
        drawer = rdMolDraw2D.MolDraw2DCairo(molSize[0], molSize[1])

    if drawOptions is None:
        drawopt = drawer.drawOptions()
        drawopt.continuousHighlight = False
    else:
        drawOptions.continuousHighlight = False
        drawer.SetDrawOptions(drawOptions)

    drawer.DrawMolecule(menv.submol, highlightAtoms=menv.highlightAtoms,
                        highlightAtomColors=menv.atomColors, highlightBonds=menv.highlightBonds,
                        highlightBondColors=menv.bondColors, highlightAtomRadii=menv.highlightRadii,
                        **kwargs)
    drawer.FinishDrawing()
    return subMol, drawer.GetDrawingText().replace('\n','')


def GetFoldedPathFragment(mol,
                          minPath=1, maxPath=7,
                          nBits=1024, maxFragment=True,
                          disposed=True):
    """Calculate folded path fragment.

    Parameters
    ----------
    mol : dkit.Chem.rdchem.Mol object
        Compound to be Calculated
    minPath : int, optional
        The probable minimum length of path-based fragment, by default 1
    maxPath : int, optional
        The probable maximum length of path-based fragment, by default 7
    nBits : int, optional
        the number of bit of morgan, by default 1014
    maxFragment : bool, optional
        Whether only return the maximum fragment of a given start atom, by default True
    disposed : bool, optional
        Whether dispose the original bitinfo, by default True

    Returns
    -------
    fragments : tuple
        The first element is the ID of all fragments generated,
        and the second one is a dict whose key is the ID of output fragments,
        value is corresponding SMARTS and svg string (is svg set as True)
    """
    bitInfo = {}
    fp = RDKFingerprint(mol,
                        minPath=minPath,
                        maxPath=maxPath,
                        fpSize=nBits,
                        bitInfo=bitInfo)

    fragments = _DisposePathFragments(
            bitInfo, maxFragment=maxFragment
        ) if disposed else bitInfo
    return fragments


def GetUnfoldedPathFragment(mol, minPath=1, maxPath=7,
                            maxFragment=True, disposed=True):
    """Calculate unfolded path fragment.

    Parameters
    ----------
    mol : dkit.Chem.rdchem.Mol object
        Compound to be Calculated
    minPath : int, optional
        The probable minimum length of path-based fragment, by default 1
    maxPath : int, optional
        The probable maximum length of path-based fragment, by default 7
    maxFragment : bool, optional
        Whether only return the maximum fragment of a given start atom, by default True
    disposed : bool, optional
        Whether dispose the original bitinfo, by default True

    Returns
    -------
    fragments : list of list
        The first element is the ID of all fragments generated
        the second one is the ID of output fragments
    """
    bitInfo = {}
    fp = UnfoldedRDKFingerprintCountBased(mol,
                                          minPath=minPath,
                                          maxPath=maxPath,
                                          bitInfo=bitInfo)

    fragments = _DisposePathFragments(
            bitInfo, maxFragment=maxFragment
        ) if disposed else bitInfo
    return fragments


def GetPathFragment(mol,
                    minPath=1, maxPath=7,
                    nBits=1024, folded=False,
                    maxFragment=True, disposed=True):
    """Calculate path fragment.

    Parameters
    ----------
    mol : dkit.Chem.rdchem.Mol object
        Compound to be Calculated
    minPath : int, optional
        The probable minimum length of path-based fragment, by default 1
    maxPath : int, optional
        The probable maximum length of path-based fragment, by default 7
    nBits : int, optional
        the number of bit of morgan, by default 1014
        this param would be ignored, if the folded set as False.
    folded : bool, optional
        which generate fragment based on unfolded fingerprint, by default True.
    maxFragment : bool, optional
        Whether only return the maximum fragment at a center atom, by default True
    disposed : bool, optional
        Whether dispose the original bitinfo, by default True

    Returns
    -------
    fragments : list of list
        The first element is the ID of all fragments generated
        the second one is the ID of output fragments
    """
    if folded:
        fragments = GetFoldedPathFragment(
            mol, minPath=minPath, maxPath=maxPath,
            nBits=nBits, maxFragment=maxFragment,
            disposed=disposed
        )
    else:
        fragments = GetUnfoldedPathFragment(
            mol, minPath=minPath, maxPath=maxPath,
            maxFragment=maxFragment, disposed=disposed
        )

    return fragments


if '__main__' == __name__:

    mol = Chem.MolFromSmiles('CN(C1=CC=C(C=C1)N=N/C2=CC=CC=C2)C')
    fragments = GetPathFragment(mol, maxFragment=True, disposed=True)
    print(fragments)



# -*- coding: utf-8 -*-
"""
Created on Sun Jun 14 17:03:20 2020

@Author: Zhi-Jiang Yang, Dong-Sheng Cao
@Institution: CBDD Group, Xiangya School of Pharmaceutical Science, CSU, China
@Homepage: http://www.scbdd.com
@Mail: yzjkid9@gmail.com; oriental-cds@163.com
@Blog: https://blog.iamkotori.com

♥I love Princess Zelda forever♥
"""


from decimal import Decimal
import multiprocessing as mp
from collections.abc import Iterable

import os
import pandas as pd
import numpy as np
import scipy as sc
import bz2
import _pickle as cPickle

from rdkit import Chem
from rdkit.Chem.Draw import rdMolDraw2D
# from IPython.display import SVG
from rdkit.Chem import rdDepictor

try:
    from .fragments import Circular, Path, FunctionGroup
except Exception:
    from fragments import Circular, Path, FunctionGroup

try:
    pd.set_option('display.max_colwidth', None)
except:
    pd.set_option('display.max_colwidth', -1)

from scipy.stats import binom


class NotFittedError(Exception):
    pass

class BaseLearner:
    """Base class for all learner in pysamsh.
    """

    def __init__(self):
        """Initialization.
        """
        self.sigFragments = None

    def GetMatrix(self, mols, **kwgrs):
        """Get the matrix of fragment of mol library.

        Parameters
        ----------
        mols : Iterable object, and each element is a rdkit.Chem.rdchem.Mol object.
            compounds, which have aspecific endpoint label, used to obtain significant fragments.
        """
    
    def ShowFragment(self, mol, **kwgrs):
        """Visulaizing fragments.

        Parameters
        ----------
        mols : Iterable object, and each element is a rdkit.Chem.rdchem.Mol object.
            compounds, which have aspecific endpoint label, used to obtain significant fragments.
        """
        pass

    def fit(self, mols,
            labels, aimLabel=1,
            minNum=5, pCutoff=0.05,
            accCutoff=None, Bonferroni=False):
        """Learning from given database which have aspecific endpoint label.

        Parameters
        ----------
        mols : Iterable object, and each element is a rdkit.Chem.rdchem.Mol object
            Compounds, which have aspecific endpoint label, used to obtain significant fragments
        labels : array-like of shape (len(mols),)
            The target values (class labels in classification)
        aimLabel : any, optional
            The label to be regarded as activity label (class labels in classification), by default 1
        minNum : int, optional
            The minimum frequency a fragment required, by default 5
        pCutoff : float, optional
            The pvalue cutoff, a fragment would be regarded as significant if its pvalue below pCutoff, by default 0.05
        accCutoff : float, optional
            The minimum accraucy lead by a fragment judge, by default None
        Bonferroni : bool, optional
            Whether use Bonferroni method to revised, by default False
        svg : bool, optional
            Whether output with a svg image, by default True

        Returns
        -------
        smash.BaseLearner
            A fitted learner, used predict() method can predict molecules without known label 
        """
        mols = tuple(mols) if isinstance(mols, Iterable) else (mols, )

        matrix = self.GetMatrix(mols, minNum=minNum)

        n = len(labels)
        m = (labels == aimLabel).sum()
        p = m/n

        sigPvalue = pd.DataFrame(
            {"ns": matrix.sum(),
             "ms": matrix[labels == 1].sum()}
        )
        sigPvalue = sigPvalue.apply(
            lambda x: sum(
                binom.pmf(range(x['ms'], x['ns']+1), x['ns'], p)
            ), axis=1
        )

        sigPvalue = sigPvalue[sigPvalue <= pCutoff]

        sigPvalue = pd.DataFrame(sigPvalue, columns=['Pvalue'])
        sigMatrix = matrix.reindex(sigPvalue.index, axis=1)
        sigPvalue['Total'] = sigMatrix.sum(axis=0)
        sigPvalue['Hitted'] = sigMatrix[labels == 1].sum(axis=0)
        sigPvalue['Accuracy'] = sigPvalue.Hitted/sigPvalue.Total
        sigPvalue['Coverage'] = sigPvalue['Hitted']/m

        if accCutoff:
            sigPvalue = sigPvalue[sigPvalue.Accuracy >= accCutoff]
        else:
            pass

        if Bonferroni:
            sigPvalue['Pvalue'] = sigPvalue.Pvalue.values * len(sigPvalue)
            sigPvalue = sigPvalue[sigPvalue.Pvalue <= pCutoff]
        else:
            pass

        substructure = {}
        for idx,vals in sigMatrix.iteritems():
            mol = mols[vals[vals==1].index[0]]
            smarts, svg = self.ShowFragment(mol, idx)
            substructure[idx] = {"SMARTS":smarts, "Substructure":svg}
        substructure = pd.DataFrame(substructure).T
        
        sigPvalue = sigPvalue.merge(substructure, left_index=True, right_index=True)
        

        sigPvalue = sigPvalue.sort_values('Pvalue')
        # print(type(sigPvalue))
        sigMatrix = sigMatrix.reindex(sigPvalue.index, axis=1)

        self.sigFragments = sigPvalue.SMARTS.to_dict()

        return sigPvalue, sigMatrix

    def predict(self, mols):
        """Predict molecules without known label 

        Parameters
        ----------
        mols : Iterable object, and each element is a rdkit.Chem.rdchem.Mol object
            Compounds to be predicted

        Returns
        -------
        y_pred : ndarray of shape (len(mols),)
            The predicted label
        predMatrix : pandas.core.frame.DataFrame
            The predictd matrix 

        Raises
        ------
        NotFittedError
            Judge a learner whether was fitted
        """
        if self.sigFragments is None:
            raise NotFittedError(
                "This instance is not fitted yet. Call 'fit' with appropriate arguments before using this method.")

        predMatrix = self.GetMatrix(mols)
        # cols = set(predMatrix.columns) & set(self.sigPvalue.index)
        predMatrix = predMatrix.reindex(
            self.sigFragments.keys(), axis=1).reset_index(drop=True).fillna(0)
        y_pred = ((predMatrix.sum(axis=1) > 0) + 0).values

        return y_pred, predMatrix

    def savePvalue(self, sigPvalue, file='./pvalue.html'):
        """Save pvalue result

        Parameters
        ----------
        file : str, optional
            The path to save result, by default './pvalue.html'
        """
        pd.set_option('colheader_justify', 'center')   # FOR TABLE <th>

        html_string = """<html>
        <head><title>HTML Pandas Dataframe with CSS</title></head>
        <style>
        /* includes alternating gray and white with on-hover color */

        .mystyle {
            font-size: 11pt; 
            font-family: Arial;
            border-collapse: collapse; 
            border: 1px solid silver;

        }

        .mystyle td, th {
            padding: 5px;
        }

        .mystyle tr:nth-child(even) {
            background: #E0E0E0;
        }

        .mystyle tr:hover {
            background: silver;
            cursor: pointer;
        }
        </style>
        <body>
            %s
        </body>
    </html>"""
        html = html_string % (sigPvalue.to_html(
            classes='mystyle', escape=False))
        with open(file, 'w') as f:
            f.write(html)
        f.close()
        return html

    def saveModel(self, file='./learner.pkl'):
        """Save fitted learner

        Parameters
        ----------
        file : str, optional
            The path to save fitted learner, by default './learner.pkl'
        """
        with bz2.BZ2File(file + '.pbz2', 'w') as f:
            cPickle.dump(self, f)
        f.close()


class CircularLearner(BaseLearner, Circular):
    """Circular fragment leanrner

    Parameters
    ----------
    BaseLearner : smash.BaseLearner
        The base learner
    Circular : smash.fragments.Circular
        The class to obtain circular fragment matrix
    """

    def __init__(self, minRadius=1, maxRadius=2,
                 nBits=1024, folded=False,
                 maxFragment=True, nJobs=1):
        """Initialization

        Parameters
        ----------
        minRadius : int, optional
            The probable minimum radius of circular fragment, by default 1
        maxRadius : int, optional
            The probable maximum radius of circular fragment, by default 2
        nBits : int, optional
            the number of bit of morgan, by default 1014
            this param would be ignored, if the folded set as False.
        folded : bool, optional
            which generate fragment based on unfolded fingerprint, by default True.
        maxFragment : bool, optional
            Whether only return the maximum fragment at a center atom, by default True
        nJobs : int, optional
            The number of CPUs to use to do the computation, by default 1
        """
        BaseLearner.__init__(self)
        Circular.__init__(self, minRadius=minRadius, maxRadius=maxRadius,
                          nBits=nBits, folded=folded,
                          maxFragment=maxFragment, nJobs=nJobs)

    def GetMatrix(self, mols, **kwgrs):
        """Rewrite method

        Parameters
        ----------
        mols : Iterable object, and each element is a rdkit.Chem.rdchem.Mol object.
            compounds, which have aspecific endpoint label, used to obtain significant fragments.

        Returns
        -------
        maxtrix : pandas.core.frame.DataFrame
            The calculated fragment matrix
        """
        matrix = self.GetCircularMatrix(mols, **kwgrs)
        return matrix

    def ShowFragment(self, mol, fragmentIndex):
        """Get the SMARTS and SVG image of circular fragment

        Parameters
        ----------
        mol : rdkit.Chem.rdchem.Mol
            The molecule which contain the aim fragment
        fragmentIndex : int
            The index of aim fragment

        Returns
        -------
        smarts : str
            The SMARTS of fragment, which could be used for screening molecules
        svg : str
            The svg string of fragment
        """ 
        smarts, svg = self.ShowCircularFragment(mol, fragmentIndex)
        return (smarts, svg)

class PathLearner(BaseLearner, Path):
    """Path-Based fragment leanrner

    Parameters
    ----------
    BaseLearner : smash.BaseLearner
        The base learner
    Path : smash.fragments.Path
        The class to obtain path-based fragment matrix
    """

    def __init__(self,
                 minPath=1, maxPath=7,
                 nBits=1024, folded=False,
                 maxFragment=True, nJobs=1):
        """Initialization

        Parameters
        ----------
        minPath : int, optional
            The probable minimum length of path-based fragment, by default 1
        maxPath : int, optional
            The probable maximum length of path-based fragment, by default 7
        nBits : int, optional
            the number of bit of morgan, by default 1014
            this param would be ignored, if the folded set as False.
        folded : bool, optional
            which generate fragment based on unfolded fingerprint, by default True.
        maxFragment : bool, optional
            Whether only return the maximum fragment at a center atom, by default True
        nJobs : int, optional
            The number of CPUs to use to do the computation, by default 1
        """
        BaseLearner.__init__(self)
        Path.__init__(self, minPath=minPath, maxPath=maxPath,
                      nBits=nBits, folded=folded, nJobs=nJobs, maxFragment=maxFragment)

    def GetMatrix(self, mols, **kwgrs):
        """Rewrite method

        Parameters
        ----------
        mols : Iterable object, and each element is a rdkit.Chem.rdchem.Mol object.
            compounds, which have aspecific endpoint label, used to obtain significant fragments.

        Returns
        -------
        maxtrix : pandas.core.frame.DataFrame
            The calculated fragment matrix
        """
        matrix = self.GetPathMatrix(mols, **kwgrs)
        return matrix

    def ShowFragment(self, mol, fragmentIndex):
        """Get the SMARTS and SVG image of Path fragment

        Parameters
        ----------
        mol : rdkit.Chem.rdchem.Mol
            The molecule which contain the aim fragment
        fragmentIndex : int
            The index of aim fragment

        Returns
        -------
        smarts : str
            The SMARTS of fragment, which could be used for screening molecules
        svg : str
            The svg string of fragment
        """     
        smarts, svg = self.ShowPathFragment(mol, fragmentIndex)
        return smarts, svg


class FunctionGroupLearner(BaseLearner, FunctionGroup):
    """Function-Group fragment leanrner

    Parameters
    ----------
    BaseLearner : smash.BaseLearner
        The base learner
    FunctionGroup : smash.fragments.FunctionGroup
        The class to obtain Function-Group fragment matrix
    """

    def __init__(self, nJobs=1):
        """Initialization

        Parameters
        ----------
        nJobs : int, optional
            The number of CPUs to use to do the computation, by default 1
        """
        BaseLearner.__init__(self)
        FunctionGroup.__init__(self, nJobs)

    def GetMatrix(self, mols, **kwgrs):
        """Rewrite method

        Parameters
        ----------
        mols : Iterable object, and each element is a rdkit.Chem.rdchem.Mol object.
            compounds, which have aspecific endpoint label, used to obtain significant fragments.

        Returns
        -------
        maxtrix : pandas.core.frame.DataFrame
            The calculated fragment matrix
        """
        matrix = self.GetFunctionGroupsMatrix(mols, **kwgrs)
        return matrix

    def ShowFragment(self, mol, fragment):

        smarts, svg = self.ShowFgFragment(mol, fragment)
        return smarts, svg

if '__main__' == __name__:
    from rdkit import Chem
    from itertools import compress
    import time

    data = pd.read_csv(r'..\tests\carc\carc.txt', sep='\t')
    # data = pd.read_csv(r'..\tests\agg\agg.csv', sep=',')
#    data = data.sample(n=10000, replace=False)

    mols = data.SMILES.map(lambda x: Chem.MolFromSmiles(x))
    bo = mols.notna().values

    mols = list(compress(mols, bo))
    y_true = data.Label.values[bo]

    # start = time.clock()
    # circular = CircularLearner(minRadius=1, maxRadius=6,
    #                            maxFragment=True, nJobs=20)

    # sigPvalue, sigMatrix = circular.fit(mols, y_true)
    # sigPvalue.to_html('././092220_1.html', escape=False)
    # print(type(sigPvalue))
    # circular.savePvalue(sigPvalue, './092220.html')
    # end = time.clock()
    # print(end-start)

    # start = time.clock()
    # pa = PathLearner(
    #     minPath=1, maxPath=7,
    #     maxFragment=True, nJobs=20
    # )

    # sigPvalue, sigMatrix = pa.fit(mols, y_true)
    # print(type(sigPvalue))
    # pa.savePvalue(sigPvalue, './092220.html')
    # end = time.clock()
    # print(end-start)

    start = time.clock()
    fg = FunctionGroupLearner(4)

    sigPvalue, sigMatrix = fg.fit(mols, y_true)
    # print(type(sigPvalue))
    fg.savePvalue(sigPvalue, './092220.html')
    print(fg.predict(mols))
    end = time.clock()
    print(end-start)


# -*- coding: utf-8 -*-
"""
Created on Sun Jun 14 15:13:34 2020

@Author: Zhi-Jiang Yang, Dong-Sheng Cao
@Institution: CBDD Group, Xiangya School of Pharmaceutical Science, CSU, China
@Homepage: http://www.scbdd.com
@Mail: yzjkid9@gmail.com; oriental-cds@163.com
@Blog: https://blog.iamkotori.com

♥I love Princess Zelda forever♥
"""


from rdkit import Chem
import pandas as pd
from rdkit.Chem.rdMolDescriptors import GetMorganFingerprintAsBitVect
from rdkit.Chem.rdMolDescriptors import GetMorganFingerprint
from rdkit.Chem.Draw import rdMolDraw2D
from rdkit.Chem.Draw import _getMorganEnv


__all__ = ['GetFoldedCircularFragment',
           'GetUnfoldedCircularFragment',
           'GetCircularFragment',
           'DrawMorganEnv']


def _DisposeCircularBitInfo(bitInfo, minRadius=1, maxFragment=True):
    """Dispose the bitinfo retrived from GetFoldedCircularFragment() or GetUnfoldedCircularFragment()
    *internal only*
    """
    allFragments = list(bitInfo.keys())

    station = {}
    # maxFragments = []
    for idx, pairs in bitInfo.items():
        for (atom, radius) in pairs:
            if radius >= minRadius and (not maxFragment or \
                atom not in station or \
                    station[atom]['radius'] < radius):
                station[atom] = {'idx':idx, 'radius':radius}

    maxFragments = [vals['idx'] for vals in station.values()]
    return (allFragments, maxFragments)


def DrawMorganEnv(mol, atomId, radius, molSize=(150, 150), baseRad=0.3, useSVG=True,
                  aromaticColor=(0.9, 0.9, 0.2), ringColor=(0.8, 0.8, 0.8),
                  centerColor=(0.6, 0.6, 0.9), extraColor=(0.9, 0.9, 0.9), drawOptions=None,
                  **kwargs):
    """Get SMARTS and SVG image from given ceten and radius
    *internal only*
    """
    menv = _getMorganEnv(mol, atomId, radius, baseRad, aromaticColor, ringColor, centerColor,
                         extraColor, **kwargs)

    submol = menv.submol
    subMol = Chem.MolToSmiles(submol, isomericSmiles=False, allBondsExplicit=True, allHsExplicit=True)
    # Drawing
    if useSVG:
        drawer = rdMolDraw2D.MolDraw2DSVG(molSize[0], molSize[1])
    else:
        drawer = rdMolDraw2D.MolDraw2DCairo(molSize[0], molSize[1])

    if drawOptions is None:
        drawopt = drawer.drawOptions()
        drawopt.continuousHighlight = False
    else:
        drawOptions.continuousHighlight = False
        drawer.SetDrawOptions(drawOptions)

    drawer.DrawMolecule(menv.submol, highlightAtoms=menv.highlightAtoms,
                        highlightAtomColors=menv.atomColors, highlightBonds=menv.highlightBonds,
                        highlightBondColors=menv.bondColors, highlightAtomRadii=menv.highlightRadii,
                        **kwargs)
    drawer.FinishDrawing()
    return subMol, drawer.GetDrawingText().replace('\n', '')


def GetFoldedCircularFragment(mol, minRadius=1, maxRadius=2,
                              nBits=1024, maxFragment=True,
                              disposed=True):
    """Get folded circular fragment

    Parameters
    ----------
    mol : dkit.Chem.rdchem.Mol object
        Compound to be Calculated
    minRadius : int, optional
        The probable minimum radius of circular fragment, by default 1
    maxRadius : int, optional
        The probable maximum radius of circular fragment, by default 2
    nBits : int, optional, 
        the number of bit of morgan, by default 1014
    maxFragment : bool, optional
        Whether only return the maximum fragment at a center atom, by default True
    disposed : bool, optional
        Whether dispose the original bitinfo, by default True

    Returns
    -------
    fragments : list of list
        The first element is the ID of all fragments generated
        the second one is the ID of output fragments
    """
    bitInfo = {}
    fp = GetMorganFingerprintAsBitVect(mol,
                                       radius=maxRadius,
                                       nBits=nBits,
                                       bitInfo=bitInfo)

    fragments = _DisposeCircularBitInfo(
            bitInfo, minRadius, maxFragment
        ) if disposed else bitInfo
    return fragments


def GetUnfoldedCircularFragment(mol, minRadius=1, maxRadius=2,
                                maxFragment=True, disposed=True):
    """Get unfolded circular fragment

    Parameters
    ----------
    mol : dkit.Chem.rdchem.Mol object
        Compound to be Calculated
    minRadius : int, optional
        The probable minimum radius of circular fragment, by default 1
    maxRadius : int, optional
        The probable maximum radius of circular fragment, by default 2
    maxFragment : bool, optional
        Whether only return the maximum fragment at a center atom, by default True
    disposed : bool, optional
        Whether dispose the original bitinfo, by default True

    Returns
    -------
    fragments : list of list
        The first element is the ID of all fragments generated
        the second one is the ID of output fragments
    """
    bitInfo = {}
    fp = GetMorganFingerprint(mol,
                              radius=maxRadius,
                              bitInfo=bitInfo)

    fragments = _DisposeCircularBitInfo(
            bitInfo, minRadius, maxFragment
        ) if disposed else bitInfo
    return fragments


def GetCircularFragment(mol, minRadius=1, maxRadius=2,
                        nBits=1024, folded=False,
                        maxFragment=True, disposed=True):
    """Get circular fragment

    Parameters
    ----------
    mol : dkit.Chem.rdchem.Mol object
        Compound to be Calculated
    minRadius : int, optional
        The probable minimum radius of circular fragment, by default 1
    maxRadius : int, optional
        The probable maximum radius of circular fragment, by default 2
    nBits : int, optional
        the number of bit of morgan, by default 1014
        this param would be ignored, if the folded set as False.
    folded : bool, optional
        which generate fragment based on unfolded fingerprint, by default True.
    maxFragment : bool, optional
        Whether only return the maximum fragment at a center atom, by default True
    nJobs : int, optional
        The number of CPUs to use to do the computation, by default 1
    disposed : bool, optional
        Whether dispose the original bitinfo, by default True

    Returns
    -------
    fragments : list of list
        The first element is the ID of all fragments generated
        the second one is the ID of output fragments
    """
    if folded:
        fragments = GetFoldedCircularFragment(mol,
                                              minRadius=minRadius, maxRadius=maxRadius,
                                              nBits=nBits, maxFragment=maxFragment,
                                              disposed=disposed)
    else:
        fragments = GetUnfoldedCircularFragment(mol,
                                                minRadius=minRadius, maxRadius=maxRadius,
                                                maxFragment=maxFragment, disposed=disposed)

    return fragments



if '__main__' == __name__:

    mol = Chem.MolFromSmiles('C1=CC2NC(=O)CC3C=2C(C(=O)C2C=CC=CC=23)=C1')

    fragments = GetCircularFragment(
        mol, maxRadius=4, minRadius=1, maxFragment=True)
    print(fragments)



# -*- encoding: utf-8 -*-
'''
Created on 2020/07/03 21:46:19

@Author: Zhi-Jiang Yang, Dong-Sheng Cao
@Institution: CBDD Group, Xiangya School of Pharmaceutical Science, CSU, China
@Homepage: http://www.scbdd.com
@Mail: yzjkid9@gmail.com; oriental-cds@163.com
@Blog: https://blog.iamkotori.com

♥I love Princess Zelda forever♥
'''


#  Original authors: Richard Hall and Guillaume Godin
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.

#
#
# Richard hall 2017
# IFG main code
# Guillaume Godin 2017
# refine output function
# astex_ifg: identify functional groups a la Ertl, J. Cheminform (2017) 9:36


from rdkit import Chem
from rdkit.Chem.Draw import rdMolDraw2D
from rdkit.Chem import rdDepictor

__all__ = [
        'GetFunctionGroupFragment',
        'DrawFgEnv'
    ]


subMolDic = {}


def DrawFgEnv(mol, subMol, molSize=(150, 150)):
    """Get the SVG image of funtion group fragment

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        The molecule which contain the aim fragment
    subMol : str
        Fragment in fragment format

    Returns
    -------
    svg : str
        The svg string of fragment
    """ 
    if not mol.GetNumConformers():
        rdDepictor.Compute2DCoords(mol)

    subMolDic[subMol] = Chem.MolFromSmarts(subMol)
    try:
        atomToUsed = mol.GetSubstructMatch(subMolDic[subMol])
        amap = {}
        submol = Chem.PathToSubmol(mol, atomToUsed, atomMap=amap)

        Chem.FastFindRings(submol)
        conf = Chem.Conformer(submol.GetNumAtoms())
        confOri = mol.GetConformer(0)
        for i1, i2 in amap.items():
            conf.SetAtomPosition(i2, confOri.GetAtomPosition(i1))
        submol.AddConformer(conf)
    except:
        submol = mol
        
    drawer = rdMolDraw2D.MolDraw2DSVG(molSize[0], molSize[1])
    drawopt = drawer.drawOptions()
    drawopt.continuousHighlight = False
    drawer.DrawMolecule(submol)
    drawer.FinishDrawing()

    svg = drawer.GetDrawingText().replace('\n', '')
    return svg


def merge(mol, marked, aset):
    """*internal only*
    """
    bset = set()
    for idx in aset:
        atom = mol.GetAtomWithIdx(idx)
        for nbr in atom.GetNeighbors():
            jdx = nbr.GetIdx()
            if jdx in marked:
                marked.remove(jdx)
                bset.add(jdx)
    if not bset:
        return
    merge(mol, marked, bset)
    aset.update(bset)


# atoms connected by non-aromatic double or triple bond to any heteroatom
# c=O should not match (see fig1, box 15).  I think using A instead of * should sort that out?
PATT_DOUBLE_TRIPLE = Chem.MolFromSmarts('A=,#[!#6]')
# atoms in non aromatic carbon-carbon double or triple bonds
PATT_CC_DOUBLE_TRIPLE = Chem.MolFromSmarts('C=,#C')
# acetal carbons, i.e. sp3 carbons connected to tow or more oxygens, nitrogens or sulfurs; these O, N or S atoms must have only single bonds
PATT_ACETAL = Chem.MolFromSmarts('[CX4](-[O,N,S])-[O,N,S]')
# all atoms in oxirane, aziridine and thiirane rings
PATT_OXIRANE_ETC = Chem.MolFromSmarts('[O,N,S]1CC1')

PATT_TUPLE = (PATT_DOUBLE_TRIPLE, PATT_CC_DOUBLE_TRIPLE,
              PATT_ACETAL, PATT_OXIRANE_ETC)


def GetFunctionGroupFragment(mol):
    """Calculate function group fragment.

    Parameters
    ----------
    mol : dkit.Chem.rdchem.Mol object
        Compound to be Calculated

    Returns
    -------
    ifgs : list
        Each element is the SMARTS of a fragment
    """    
    marked = set()
# mark all heteroatoms in a molecule, including halogens
    for atom in mol.GetAtoms():
        if atom.GetAtomicNum() not in (6, 1):  # would we ever have hydrogen?
            marked.add(atom.GetIdx())

# mark the four specific types of carbon atom
    for patt in PATT_TUPLE:
        for path in mol.GetSubstructMatches(patt):
            for atomindex in path:
                marked.add(atomindex)

# merge all connected marked atoms to a single FG
    groups = []
    while marked:
        grp = set([marked.pop()])
        merge(mol, marked, grp)
        groups.append(grp)

# extract also connected unmarked carbon atoms
    ifgs = []
    st = []
    for g in groups:
        uca = set()
        for atomidx in g:
            for n in mol.GetAtomWithIdx(atomidx).GetNeighbors():
                if n.GetAtomicNum() == 6:
                    uca.add(n.GetIdx())
        fg = Chem.MolFragmentToSmiles(mol, g.union(uca), canonical=True)
        
        if fg not in st:
            ifgs.append(fg)
            st.append(fg)
    
    return ifgs


if __name__ == "__main__":
    from rdkit import Chem
    
    mol = Chem.MolFromSmiles('CN(C1=CC=C(C=C1)N=N/C2=CC=CC=C2)C')
    fragments = GetFunctionGroupFragment(mol)
    print(fragments)



# -*- coding: utf-8 -*-
"""
Created on Sun Jun 14 15:12:50 2020

@Author: Zhi-Jiang Yang, Dong-Sheng Cao
@Institution: CBDD Group, Xiangya School of Pharmaceutical Science, CSU, China
@Homepage: http://www.scbdd.com
@Mail: yzjkid9@gmail.com; oriental-cds@163.com
@Blog: https://blog.iamkotori.com

♥I love Princess Zelda forever♥
"""


from collections import Counter
from functools import partial
from collections.abc import Iterable
from multiprocessing import Pool
from multiprocessing import freeze_support
import numpy as np
import pandas as pd
from rdkit import Chem

try:
    from .path import *
    from .circular import *
    from .functionGroup import *
except:
    from path import *
    from circular import *
    from functionGroup import *


class Circular:
    """Calculte circular fragments and obtain matrix  
    """    
    def __init__(self, minRadius=1, maxRadius=2, 
                 nBits=1024, folded=False,
                 maxFragment=True, nJobs=1):
        """Initialization

        Parameters
        ----------
        minRadius : int, optional
            The probable minimum radius of circular fragment, by default 1
        maxRadius : int, optional
            The probable maximum radius of circular fragment, by default 2
        nBits : int, optional
            the number of bit of morgan, by default 1014
            this param would be ignored, if the folded set as False.
        folded : bool, optional
            which generate fragment based on unfolded fingerprint, by default True.
        maxFragment : bool, optional
            Whether only return the maximum fragment at a center atom, by default True
        nJobs : int, optional
            The number of CPUs to use to do the computation, by default 1
        """
        self.nBits = nBits if folded else None
        self.folded = folded
        self.maxRadius = maxRadius
        self.minRadius = minRadius
        self.maxFragment = maxFragment
        self.nJobs = nJobs if nJobs >= 1 else None

    def GetCircularFragmentLib(self, mols):
        """Calculate circular fragments

        Parameters
        ----------
        mols : Iterable object, and each element is a rdkit.Chem.rdchem.Mol object.
            The compounds used to obtain circular fragments.
        svg : bool, optional
            Whether output with a svg image, by default False

        Returns
        -------
        fragments : list
            Fragments calculated, each element is a list of list
            return from GetCircularFragment() function
        """        
        func = partial(GetCircularFragment,
                       minRadius=self.minRadius,
                       maxRadius=self.maxRadius,
                       nBits=self.nBits,
                       folded=self.folded,
                       maxFragment=self.maxFragment,
                       disposed=True)

        mols = mols if isinstance(mols, Iterable) else (mols, )

        pool = Pool(self.nJobs)
        fragments = pool.map_async(func, mols).get()
        pool.close()
        pool.join()

        return fragments
    
    def _getIdx(self, subArray, Array):
        """Get index of a list to another list
        *internal only*
        """
        uni = set(Array)&set(subArray)
        idx = [Array.index(x) for x in uni]
        return idx
    
    def GetCircularMatrix(self, mols, minNum=None):
        """return circular matrix
    
        Parameters
        ----------
        mols : Iterable object, and each element is a rdkit.Chem.rdchem.Mol object
            The compounds used to obtain circular fragments matrix
        minNum : bool, optional
            Whether output with a svg image, by default False

        Returns
        -------
        pandas.core.frame.DataFrame
            The fragment matrix of molecules
        """

        fragments = self.GetCircularFragmentLib(mols)
        unique = [frag for allFragments in fragments for frag in allFragments[1]]
        
        if minNum:
            _all = [bit for info in fragments for bit in info[0]]
            c = Counter(_all)
            unique = [x for x in unique if c[x] >= minNum]
            
            
        unique = list(set(unique))

        num = len(unique)
        matrix = np.zeros((len(mols), num), dtype=np.int8)
        
        fragments = pd.DataFrame(fragments)
        subArray = fragments[0]
        func = partial(self._getIdx, Array=unique)
        pool = Pool(self.nJobs)
        st = pool.map_async(func, subArray).get()
        pool.close()
        pool.join()
        
        for idx,arr in zip(st, matrix):
            arr[idx]=1
        matrix = pd.DataFrame(matrix, columns=unique)
         
        return matrix

    def ShowCircularFragment(self, mol, fragmentIndex):
        """Get the SMARTS and SVG image of circular fragment

        Parameters
        ----------
        mol : rdkit.Chem.rdchem.Mol
            The molecule which contain the aim fragment
        fragmentIndex : int
            The index of aim fragment

        Returns
        -------
        smarts : str
            The SMARTS of fragment, which could be used for screening molecules
        svg : str
            The svg string of fragment
        """ 
        fragments = GetCircularFragment(
                mol, minRadius=self.minRadius, maxRadius=self.maxRadius,
                nBits=self.nBits, folded=self.folded, maxFragment=self.maxFragment,
                disposed=False
            )
        center, radius = fragments[fragmentIndex][0]
        smarts, svg = DrawMorganEnv(mol, center, radius)
        return (smarts, svg)
        
        
class Path:
    """Calculte path-based fragments and obtain matrix  
    """    
    def __init__(self,
                 minPath=1, maxPath=7,
                 nBits=1024, folded=False,
                 maxFragment=True, nJobs=1):
        
        """Initialization

        Parameters
        ----------
        minPath : int, optional
            The probable minimum length of path-based fragment, by default 1
        maxPath : int, optional
            The probable maximum length of path-based fragment, by default 7
        nBits : int, optional
            the number of bit of morgan, by default 1014
            this param would be ignored, if the folded set as False.
        folded : bool, optional
            which generate fragment based on unfolded fingerprint, by default True.
        maxFragment : bool, optional
            Whether only return the maximum fragment at a center atom, by default True
        nJobs : int, optional
            The number of CPUs to use to do the computation, by default 1
        """
        self.minPath = minPath
        self.maxPath = maxPath
        self.nBits = nBits if folded else None
        self.folded = folded
        self.nJobs = nJobs if nJobs >= 1 else None
        self.maxFragment = maxFragment


    def GetPathFragmentLib(self, mols):
        """Calculate path-based fragments

        Parameters
        ----------
        mols : Iterable object, and each element is a rdkit.Chem.rdchem.Mol object
            The compounds used to obtain path-based fragment

        Returns
        -------
        fragments : list
            Fragments calculated, each element is returned from GetPathFragment() function
        """
        mols = mols if isinstance(mols, Iterable) else (mols,)
        func = partial(GetPathFragment,
                       minPath=self.minPath,
                       maxPath=self.maxPath,
                       nBits=self.nBits,
                       folded=self.folded,
                       maxFragment=self.maxFragment,
                       disposed=True)

        pool = Pool(self.nJobs)
        fragments = pool.map_async(func, mols).get()
        pool.close()
        pool.join()

        return fragments
    
    def _getIdx(self, subArray, Array):
        """Get index of a list to another list
        *internal only*
        """
        uni = set(Array) & set(subArray)
        idx = [Array.index(x) for x in uni]
        return idx
    
    def GetPathMatrix(self, mols, minNum=None):
        """return path-based matrix

        Parameters
        ----------
        mols : Iterable object, and each element is a rdkit.Chem.rdchem.Mol object
            The compounds used to obtain path-based matrix
        minNum : bool, optional
            Whether output with a svg image, by default False

        Returns
        -------
        matrix : pandas.core.frame.DataFrame
            The fragment matrix of molecules
        """
        fragments = self.GetPathFragmentLib(mols)
        
        unique = [bit for info in fragments for bit in info[1]]

        if minNum:
            _all = [bit for info in fragments for bit in info[0]]
            c = Counter(_all)
            unique = [x for x in unique if c[x] >= minNum]
            
            
        unique = list(set(unique))

        num = len(unique)
        matrix = np.zeros((len(mols), num), dtype=np.int8)
    
        fragments = pd.DataFrame(fragments)
        sub = fragments[0]
        func = partial(self._getIdx, Array=unique)
        
        pool = Pool(self.nJobs)
        st = pool.map_async(func, sub).get()
        pool.close()
        pool.join()
        for idx,arr in zip(st, matrix):
            arr[idx]=1
        matrix = pd.DataFrame(matrix, columns=unique)
         
        return matrix

    def ShowPathFragment(self, mol, fragmentIndex):
        """Get the SMARTS and SVG image of Path fragment

        Parameters
        ----------
        mol : rdkit.Chem.rdchem.Mol
            The molecule which contain the aim fragment
        fragmentIndex : int
            The index of aim fragment

        Returns
        -------
        smarts : str
            The SMARTS of fragment, which could be used for screening molecules
        svg : str
            The svg string of fragment
        """        
        fragments = GetPathFragment(
                mol, minPath=self.minPath, maxPath=self.maxPath,
                nBits=self.nBits, folded=self.folded, maxFragment=self.maxFragment,
                disposed=False
            )
        
        bondPath = fragments[fragmentIndex][0]
        try:
            smarts, svg = DrawRDKitEnv(mol, bondPath)
        except Chem.KekulizeException:
            kSmi = Chem.MolToSmiles(mol, kekuleSmiles=True)
            mol = Chem.MolFromSmarts(kSmi)
            smarts, svg = DrawRDKitEnv(mol, bondPath)
        return smarts, svg
    

class FunctionGroup:
    """Calculte function-group fragments and obtain matrix  
    """    
    def __init__(self, nJobs=1):
        """Initialization

        Parameters
        ----------
        nJobs : int, optional
            The number of CPUs to use to do the computation, by default 1
        """
        self.nJobs = nJobs if nJobs > 0 else None

    def GetFunctionGroupFragmentLib(self, mols):
        """Calculate function-group fragments

        Parameters
        ----------
        mols : Iterable object, and each element is a rdkit.Chem.rdchem.Mol object
            The compounds used to obtain function-group fragment

        Returns
        -------
        fragments : list
            Fragments calculated, each element is rerurned from GetFunctionGroupFragment() function
        """
        mols = mols if isinstance(mols, Iterable) else (mols,)
        func = partial(GetFunctionGroupFragment)
        pool = Pool(self.nJobs)
        fgs = pool.map_async(func, mols).get()
        pool.close()
        pool.join()
        return fgs
     
    def _getIdx(self, subArray, Array):
        """
        """
        uni = set(Array) & set(subArray)
        idx = [Array.index(x) for x in uni]
#        print(idx)
        return idx
    
    def GetFunctionGroupsMatrix(self, mols, minNum=None):
        """return function-group matrix

        Parameters
        ----------
        mols : Iterable object, and each element is a rdkit.Chem.rdchem.Mol object
            The compounds used to obtain function-group matrix

        Returns
        -------
        pandas.core.frame.DataFrame
            the fragment matrix of molecules
        """
        
        fgs = self.GetFunctionGroupFragmentLib(mols)
#        return fgs
#        print(fgs)
    
        unique = [x for fg in fgs for x in fg]
        if minNum:
            c = Counter(unique)
            unique = [x for x in unique if c[x] >= minNum]
        unique = list(set(unique))
        
        num = len(unique)
        matrix = np.zeros((len(mols), num), dtype=np.int8)
        
#        cols =  ['SMARTS'] if not svg else ['SMARTS', 'Substructure']
#        fgs = pd.DataFrame([fg for fg in sum(fgs,[])], columns=cols)
        
#        fgs = pd.DataFrame(fgs)
        # sub = [[x[0] for x in fg] for fg in fgs]
        func = partial(self._getIdx, Array=unique)
        
        pool = Pool(self.nJobs)
        st = pool.map_async(func, fgs).get()
        pool.close()
        pool.join()
#        st = st.apply(lambda x: [dic[i] for i in x])
        for idx,arr in zip(st, matrix):
            arr[idx]=1
        matrix = pd.DataFrame(matrix, columns=unique)

        return matrix

    def ShowFgFragment(self, mol, fragment):
        """Get the SMARTS and SVG image of functional group fragment

        Parameters
        ----------
        mol : rdkit.Chem.rdchem.Mol
            The molecule which contain the aim fragment
        fragment : int
            The fragment in SMARTS format

        Returns
        -------
        smarts : str
            The SMARTS of fragment, which could be used for screening molecules
        svg : str
            The svg string of fragment
        """ 
        svg = DrawFgEnv(mol, fragment)
        smarts = fragment
        return smarts, svg

if '__main__' == __name__:
    freeze_support()
    from rdkit import Chem
    import time
    
    data = pd.read_csv(r'..\..\tests\Carc\Carc.txt', sep='\t')
    # data = data.sample(frac=0.2)
    mols = data.SMILES.map(lambda x: Chem.MolFromSmiles(x))
    # mols = [Chem.MolFromSmiles(smi) for smi in smis]

    # start = time.clock()
    # circular = Circular(folded=False, maxRadius=7,
    #                     minRadius=1, maxFragment=True,
    #                     nJobs=20)
    # # cirMatrix = circular.GetCircularMatrix(mols, minNum=5)
    # svg = circular.ShowCircularFragment(mols[1115], 2380084179)
    # print(svg)
    # end = time.clock()
    # print(end - start)
    # circular_matrix = circular.GetCircularMatrix(mols, svg=True)

    # path = Path(minPath=1, maxPath=4)
    # # pathMatrix = path.GetPathMatrix(mols)
    # # print(pathMatrix)
    # svg = path.ShowPathFragment(mols[29], 1003675631)
    # print(svg)
    # print(path.substructure)
    # path_matrix = path.GetPathMatrix()
    # print(pathMatrix.shape)

    funcgroup = FunctionGroup()
    print(funcgroup.GetFunctionGroupsMatrix(mols, minNum=5))
    print('Done!!')


# -*- coding: utf-8 -*-
"""
Created on Fri Mar 15 13:21:07 2020

@Author: Zhi-Jiang Yang, Dong-Sheng Cao
@Institution: CBDD Group, Xiangya School of Pharmaceutical Science, CSU, China
@Homepage: http://www.scbdd.com
@Mail: yzjkid9@gmail.com; oriental-cds@163.com
@Blog: https://blog.moyule.me

I love my senpai forerver!:P
"""

from __future__ import absolute_import
from distutils.core import setup

# print（__doc__）

package_data= {'smash':['fragments/*']}


setup(name="smasher",  
      version="1.0",
      license="MIT",
      description="",
      long_description="PySmash, a user-friendly and integrated tool which offers convenience and advantage for the automatic structural alerts derivation and application is presented. The current version of PySmash provides both software and Python package, which achieves both flexibility and ease of pipeline integration. Three kinds of substructure derivation algorithm, including circular, path-based and function group-based algorithms are provided, where researchers can customize own parameters about substructure size, accuracy and coverage, statistical significance and server deployment. Besides, PySmash also provides prediction function for derived substructure application.",
      author="Zhi-Jiang Yang (Kotori), Dong-Sheng Cao",
      author_email="yzjkid9@gmail.com",
      maintainer="Zhi-Jiang Yang (Kotori)",
      maintainer_email="kotori@cbdd.me",
      url="https://github.com/kotori-y/pysmash",
      package_data=package_data,
      # include_package_data=True,
      package_dir={'smash':'smash'},
      #py_modules = ['scopy.ScoConfig'],
      packages=['smash']
      )
